@@ -3,6 +3,7 @@ use std::net::SocketAddr;
 use std::sync::Arc;
 
 use futures::future::Future;
+use hyper::header;
 use hyper::server::conn::AddrStream;
 use hyper::service::{make_service_fn, service_fn};
 use hyper::{Body, Method, Request, Response, Server};
@@ -14,6 +15,7 @@ use garage_model::garage::Garage;
 use crate::error::*;
 use crate::signature::check_signature;
 
+use crate::helpers::*;
 use crate::s3_bucket::*;
 use crate::s3_copy::*;
 use crate::s3_delete::*;
@@ -82,11 +84,27 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 	let path = req.uri().path().to_string();
 	let path = percent_encoding::percent_decode_str(&path).decode_utf8()?;
 	let (api_key, content_sha256) = check_signature(&garage, &req).await?;
-	if path == "/" {
+
+	let authority = req
+		.headers()
+		.get(header::HOST)
+		.ok_or_else(|| Error::BadRequest("HOST header required".to_owned()))?
+		.to_str()?;
+
+	let host = authority_to_host(authority)?;
+
+	let bucket = garage
+		.config
+		.s3_api
+		.root_domain
+		.as_ref()
+		.and_then(|root_domain| host_to_bucket(&host, root_domain));
+
+	if path == "/" && bucket.is_none() {
 		return handle_list_buckets(&api_key);
 	}
 
-	let (bucket, key) = parse_bucket_key(&path)?;
+	let (bucket, key) = parse_bucket_key(&path, bucket)?;
 	let allowed = match req.method() {
 		&Method::HEAD | &Method::GET => api_key.allow_read(bucket),
 		_ => api_key.allow_write(bucket),
@@ -137,7 +155,7 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 					let copy_source = req.headers().get("x-amz-copy-source").unwrap().to_str()?;
 					let copy_source =
 						percent_encoding::percent_decode_str(copy_source).decode_utf8()?;
-					let (source_bucket, source_key) = parse_bucket_key(&copy_source)?;
+					let (source_bucket, source_key) = parse_bucket_key(&copy_source, None)?;
 					if !api_key.allow_read(source_bucket) {
 						return Err(Error::Forbidden(format!(
 							"Reading from bucket {} not allowed for this key",
@@ -245,13 +263,25 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 	}
 }
 
-/// Extract the bucket name and the key name from an HTTP path
+/// Extract the bucket name and the key name from an HTTP path and possibly a bucket provided in
+/// the host header of the request
 ///
 /// S3 internally manages only buckets and keys. This function splits
 /// an HTTP path to get the corresponding bucket name and key.
-fn parse_bucket_key(path: &str) -> Result<(&str, Option<&str>), Error> {
+fn parse_bucket_key<'a>(
+	path: &'a str,
+	host_bucket: Option<&'a str>,
+) -> Result<(&'a str, Option<&'a str>), Error> {
 	let path = path.trim_start_matches('/');
 
+	if let Some(bucket) = host_bucket {
+		if !path.is_empty() {
+			return Ok((bucket, Some(path)));
+		} else {
+			return Ok((bucket, None));
+		}
+	}
+
 	let (bucket, key) = match path.find('/') {
 		Some(i) => {
 			let key = &path[i + 1..];
@@ -275,7 +305,7 @@ mod tests {
 
 	#[test]
 	fn parse_bucket_containing_a_key() -> Result<(), Error> {
-		let (bucket, key) = parse_bucket_key("/my_bucket/a/super/file.jpg")?;
+		let (bucket, key) = parse_bucket_key("/my_bucket/a/super/file.jpg", None)?;
 		assert_eq!(bucket, "my_bucket");
 		assert_eq!(key.expect("key must be set"), "a/super/file.jpg");
 		Ok(())
@@ -283,10 +313,10 @@ mod tests {
 
 	#[test]
 	fn parse_bucket_containing_no_key() -> Result<(), Error> {
-		let (bucket, key) = parse_bucket_key("/my_bucket/")?;
+		let (bucket, key) = parse_bucket_key("/my_bucket/", None)?;
 		assert_eq!(bucket, "my_bucket");
 		assert!(key.is_none());
-		let (bucket, key) = parse_bucket_key("/my_bucket")?;
+		let (bucket, key) = parse_bucket_key("/my_bucket", None)?;
 		assert_eq!(bucket, "my_bucket");
 		assert!(key.is_none());
 		Ok(())
@@ -294,11 +324,30 @@ mod tests {
 
 	#[test]
 	fn parse_bucket_containing_no_bucket() {
-		let parsed = parse_bucket_key("");
+		let parsed = parse_bucket_key("", None);
 		assert!(parsed.is_err());
-		let parsed = parse_bucket_key("/");
+		let parsed = parse_bucket_key("/", None);
 		assert!(parsed.is_err());
-		let parsed = parse_bucket_key("////");
+		let parsed = parse_bucket_key("////", None);
 		assert!(parsed.is_err());
 	}
+
+	#[test]
+	fn parse_bucket_with_vhost_and_key() -> Result<(), Error> {
+		let (bucket, key) = parse_bucket_key("/a/super/file.jpg", Some("my-bucket"))?;
+		assert_eq!(bucket, "my-bucket");
+		assert_eq!(key.expect("key must be set"), "a/super/file.jpg");
+		Ok(())
+	}
+
+	#[test]
+	fn parse_bucket_with_vhost_no_key() -> Result<(), Error> {
+		let (bucket, key) = parse_bucket_key("", Some("my-bucket"))?;
+		assert_eq!(bucket, "my-bucket");
+		assert!(key.is_none());
+		let (bucket, key) = parse_bucket_key("/", Some("my-bucket"))?;
+		assert_eq!(bucket, "my-bucket");
+		assert!(key.is_none());
+		Ok(())
+	}
 }
